#include "digest.h"

#define BUF_MAXSIZE 4096

//=========================
// PRIVATE
//=========================

/**
 * @brief Stringify the given bytes digest
 *
 * @param md_len
 * @param md_value
 * @return char*
 */
static char *stringify_digest(unsigned int md_len, unsigned char md_value[]) {
    size_t hexlen = 2;
    size_t outdigestlen = md_len * hexlen;

    char *digest = malloc(outdigestlen + 1);
    char *hex = digest;

    for (size_t i = 0; i < md_len; i++) {
        hex += sprintf(hex, "%02x", md_value[i]);
    }

    return digest;
}

/**
 * @brief Initialize the EVP context
 *
 * @param digest_method
 * @return EVP_MD_CTX*
 */
EVP_MD_CTX *EVP_context_init(char *digest_method) {
    EVP_MD_CTX *mdctx;
    const EVP_MD *md;

    if (digest_method == NULL) {
        fprintf(stderr, "Digest method not provided.\n");
        exit(ENOMSG);
    }

    md = EVP_get_digestbyname(digest_method);
    if (md == NULL) {
        fprintf(stderr, "Digest method not found.\n");
        exit(ENOPROTOOPT);
    }

    mdctx = EVP_MD_CTX_new();
    EVP_DigestInit_ex2(mdctx, md, NULL);

    return mdctx;
}

//=========================
// PUBLIC
//=========================

char *str_digest(char *foo, char *digest_method) {
    EVP_MD_CTX *mdctx = EVP_context_init(digest_method);
    unsigned char md_value[EVP_MAX_MD_SIZE];
    unsigned int md_len;

    EVP_DigestUpdate(mdctx, foo, strlen(foo));
    EVP_DigestFinal_ex(mdctx, md_value, &md_len);
    EVP_MD_CTX_free(mdctx);

    return stringify_digest(md_len, md_value);
}

char *file_digest(char *filename, char *digest_method) {
    EVP_MD_CTX *mdctx = EVP_context_init(digest_method);
    unsigned char md_value[EVP_MAX_MD_SIZE];
    unsigned int md_len;

    FILE *f;
    char buf[BUF_MAXSIZE];

    // case of STDIN
    if (filename == NULL) {
        f = stdin;
    } else {
        if ((f = fopen(filename, "r")) == NULL) {
            perror(filename);
            exit(ENOENT);
        }
    }

    while (fgets(buf, BUF_MAXSIZE, f) != NULL) {
        EVP_DigestUpdate(mdctx, buf, strlen(buf));
    }

    EVP_DigestFinal_ex(mdctx, md_value, &md_len);
    EVP_MD_CTX_free(mdctx);

    return stringify_digest(md_len, md_value);
}
